/*
 * Beaver.cpp
 *
 */

#ifndef PROTOCOLS_BEAVER_HPP_
#define PROTOCOLS_BEAVER_HPP_

#include "Beaver.h"

#include "Replicated.hpp"

#include <array>

#include <optional>
#include <numeric>

template<typename T, typename I>
T pow(T x, I exponent)
{
    assert(exponent >= I{0});

    if (exponent == I{0})
        return T{1};

    auto result = T{1};

    while (exponent > I{0})
    {
        if ((exponent & I{1}) != I{0})
        {
            result *= x;
        }
        x *= x;

        exponent >>= 1;
    }

    return result;
}

template<class T>
Player& Beaver<T>::branch()
{
    return P;
}

template<class T>
void Beaver<T>::init_mul(SubProcessor<T>* proc)
{
    assert(proc != 0);
    init_mul(proc->DataF, proc->MC);
}

template<class T>
void Beaver<T>::init_mul(Preprocessing<T>& prep, typename T::MAC_Check& MC)
{
    this->prep = &prep;
    this->MC = &MC;
    shares.clear();
    opened.clear();
    previously_opened.clear();
    triples.clear();
}

template<class T>
typename T::clear Beaver<T>::prepare_mul(const T& x, const T& y, int n)
{
    (void) n;
    triples.push_back({{}});
    auto& triple = triples.back();
    triple = prep->get_triple(n);
    shares.push_back(x - triple[0]);
    shares.push_back(y - triple[1]);
    return 0;
}

template<typename T>
consteval T binom(T n, T k)
{
    if (k == T{0} or k == n)
        return T{1};
    else
        return binom(n - T{1}, k) + binom(n - T{1}, k - T{1});
}

template<typename T, std::size_t n>
consteval std::array<T, n + T{1}> binom_row()
{
    std::array<T, n + T{1}> result;
    for (auto k = T{0}; k <= T{n}; ++k)
    {
        result[k] = binom<T>(n, k);
    }
    return result;
}

template<typename T, std::size_t n>
consteval auto binom_triangle()
{
    return []<std::size_t... k>(std::index_sequence<k...>)
    {
        return std::make_tuple(binom_row<T, k>()...);
    }(std::make_index_sequence<n + 1>{});
}



template<class T>
void Beaver<T>::adjust_binomial_coefficients(int m)
{
    auto old_size = binomial_coefficients.size();
    if (m < 0)
    {
        binomial_coefficients.resize(prepared_binomial_coefficients + 1);
        if (old_size < binomial_coefficients.size())
        {
            [this]<std::size_t... k>(std::index_sequence<k...>)
            {
                constexpr auto triangle = binom_triangle<int, prepared_binomial_coefficients>();
                ((binomial_coefficients[k] = std::vector<int>(std::get<k>(triangle).begin(), std::get<k>(triangle).end())), ...);
            }(std::make_index_sequence<prepared_binomial_coefficients + 1>{});
        }
    }
    else if (static_cast<std::size_t>(m) >= binomial_coefficients.size())
    {
        binomial_coefficients.resize(m + 1);
        for (int n = old_size; n < m + 1; ++n)
        {
            binomial_coefficients[n] = std::vector<int>(n + 1);
            binomial_coefficients[n][0] = 1;
            binomial_coefficients[n][n] = 1;
            for (int k = 1; k < n; ++k)
            {
                binomial_coefficients[n][k] = binomial_coefficients[n - 1][k] + binomial_coefficients[n - 1][k - 1];
            }
        }
    }
}

template<class T>
void Beaver<T>::prepare_power(T variable, int exponent)
{
    (void)exponent;

    auto mask = T{};

    shares.push_back(variable - mask);
}

template<class T>
T Beaver<T>::finalize_power(int exponent)
{
    if (exponent == 0)
        return T::constant(1, P.my_num(), MC->get_alphai());

    adjust_binomial_coefficients(exponent);

    auto tuple = std::vector<T>(exponent + 1, T{});
    tuple[0] = T::constant(1, P.my_num(), MC->get_alphai());

    auto y = *it;

    auto result = T{};
    for (int k = 0; k <= exponent; ++k)
    {
        result += tuple[k] * typename T::clear{binomial_coefficients[exponent][k]} * ::pow(y, exponent - k);
    }

    ++it;

    return result;
}

template<typename T>
void Beaver<T>::finalize_prefix_power(register_span<T> result)
{
    int exponent = result.size();

    assert(exponent > 0);

    adjust_binomial_coefficients(exponent);

    auto tuple = std::vector<T>(exponent + 1, T{});
    tuple[0] = T::constant(1, P.my_num(), MC->get_alphai());

    auto y = *it;

    for (int e = 1; e <= exponent; ++e)
    {
        auto z = T{};
        for (int k = 0; k <= e; ++k)
        {
            z += tuple[k] * typename T::clear{binomial_coefficients[e][k]} * ::pow(y, e - k);
        }
        result[e - 1] = z;
    }

    ++it;
}

template<class T>
void Beaver<T>::prepare_product(register_span<T const> factors)
{
    auto masks = std::vector<T>(factors.size(), T{});

    for (std::size_t i = 0; i < factors.size(); ++i)
    {
        shares.push_back(factors[i] - masks[i]);
    }
}

constexpr std::size_t MAX_BASECASE_SIZE = 3;

auto split_indices(std::span<int const> indices)
{
    auto n = indices.size();
    auto split = n - n/2;
    // auto other_partition = n/2;
    // if (split & other_partition & 1) // both odd
    // {
    //     split -= 1;
    //     assert(split > 0);
    // }
    return std::make_pair(indices.first(split), indices.subspan(split));
}

template<typename T>
constexpr T pow2(T exponent) noexcept
{
    return T{1} << exponent;
}

template<typename T>
T& next_tuple_element(std::vector<T>& tuple) /*TODO: refactor to take a vector<T const>::iterator instead*/
{
    return tuple.emplace_back();
}

template<bool WithPrefactor, typename T>
void add_tuple(T one, auto opened, auto tuple_element, std::size_t count, T& sum)
{
    if constexpr (not WithPrefactor)
    {
        auto product = one;
        for (std::size_t i = 0; i < count; ++i)
        {
            product *= opened(i);
        }
        sum += product;
    }

    assert(count < static_cast<std::size_t>(std::numeric_limits<std::size_t>::digits));

    for (std::size_t i = not WithPrefactor; i < ::pow2(count); ++i)
    {
        T product = tuple_element(i);
        for (std::size_t j = 0; j < count; ++j)
        {
            if ((::pow2(j) & i) == 0)
            {
                product *= opened(j);
            }
        }
        sum += product;
    }
}

template<bool WithPrefactor, typename T>
void continue_helper_add_tuple(T one, auto opened, std::vector<T>& tuple, std::span<int const> indices, T& sum)
{
    add_tuple<WithPrefactor>(one, [indices, opened](std::size_t i){ return opened[indices[i]]; }, [&tuple](std::size_t){ return next_tuple_element(tuple); }, indices.size(), sum);
}

template<typename T>
void continue_helper2(T one, auto opened, std::vector<T>& tuple /*TODO: refactor to take a vector<T const>::iterator instead*/, std::span<int const> indices, T& sum)
{
    if (indices.size() <= MAX_BASECASE_SIZE)
    {
        continue_helper_add_tuple<true>(one, opened, tuple, indices, sum);
        return;
    }

    auto [bot, top] = split_indices(indices);

    continue_helper2(one, opened, tuple, bot, sum);
    continue_helper2(one, opened, tuple, top, sum);
}

template<bool IsAdditive, typename T>
void continue_helper1(T one, auto&& open, auto opened, std::vector<T>& tuple /*TODO: refactor to take a vector<T const>::iterator instead*/, std::span<int const> indices, T& sum)
{
    if (indices.size() <= MAX_BASECASE_SIZE)
    {
        continue_helper_add_tuple<true>(one, opened, tuple, indices, sum);
        return;
    }

    auto [bot, top] = split_indices(indices);
    if constexpr (IsAdditive)
    {
        T right;
        continue_helper1<false>(one, open, opened, tuple, top, right);
        open(right);
        continue_helper1<true>(one, open, opened, tuple, bot, sum);
        continue_helper2(one, opened, tuple, top, sum);
    }
    else
    {
        T left;
        continue_helper1<false>(one, open, opened, tuple, bot, left);
        open(left);
        continue_helper2(one, opened, tuple, bot, sum);
        continue_helper1<true>(one, open, opened, tuple, top, sum);
    }
}

template<typename T>
void continue_helper0(T one, auto&& open, auto opened, std::vector<T>& tuple /*TODO: refactor to take a vector<T const>::iterator instead*/, std::span<int const> indices, T& sum)
{
    if (indices.size() <= MAX_BASECASE_SIZE)
    {
        continue_helper_add_tuple<false>(one, opened, tuple, indices, sum);
        return;
    }

    auto [bot, top] = split_indices(indices);
    T left;
    continue_helper0(one, open, opened, tuple, bot, left);
    open(left);
    T right;
    continue_helper0(one, open, opened, tuple, top, right);
    open(right);

    continue_helper1<false>(one, open, opened, tuple, bot, sum);
    continue_helper1<false>(one, open, opened, tuple, top, sum);
}

template<class T>
void Beaver<T>::continue_product(std::size_t factor_count)
{
    std::vector<T> tuple;

    auto open = [this](T value)
    {
        shares.push_back(value);
    };

    std::vector<int> indices(factor_count);
    std::iota(begin(indices), end(indices), 0);

    T last_unopened_element; // TODO: properly store/load this value
    T one = T::constant(1, P.my_num(), MC->get_alphai());
    continue_helper0(one, open, it, tuple, indices, last_unopened_element);

    it += factor_count;
}

template<typename T>
struct helper_node
{
    std::optional<T> value;
    bool is_additive = false;
    int left = -1;
    int right = -1;
    int additive_left = -1;
    int additive_right = -1;
};

template<typename T>
using helper_tree = std::vector<helper_node<T>>;

template<typename T, typename... Args>
int add_to_tree(helper_tree<T>& tree, Args&&... args)
{
    auto n = tree.size();
    tree.emplace_back(std::forward<Args>(args)...);
    return n;
}

template<typename T>
void finalize_helper2(helper_tree<T>& tree, std::span<int const> indices, T& sum, int index0, int index1)
{
    if (indices.size() <= MAX_BASECASE_SIZE)
    {
        return;
    }

    auto y0 = tree.at(index0);
    auto y0_next_index = y0.is_additive ? y0.right : y0.left;
    auto y0_next = tree.at(y0_next_index);
    auto y0_next_additive_index = y0.is_additive ? y0.additive_left : y0.additive_right;

    auto y1 = tree.at(index1);
    auto y1_next_index = y1.is_additive ? y1.right : y1.left;
    auto y1_next = tree.at(y1_next_index);
    auto y1_next_additive_index = y1.is_additive ? y1.additive_left : y0.additive_right;

    sum += y0_next.value.value() * y1_next.value.value();

    auto [bot, top] = split_indices(indices);

    finalize_helper2(tree, bot, sum, y0_next_index, y1_next_additive_index);
    finalize_helper2(tree, top, sum, y1_next_index, y0_next_additive_index);
}

template<bool IsAdditive, typename T>
int finalize_helper1(auto next_opened, helper_tree<T>& tree, std::span<int const> indices, T& sum, int index0)
{
    if (indices.size() <= MAX_BASECASE_SIZE)
    {
        if constexpr (IsAdditive)
        {
            return -1;
        }
        else
        {
            auto opened = next_opened();
            return add_to_tree(tree, opened, IsAdditive);
        }
    }

    auto [bot, top] = split_indices(indices);
    
    auto node0 = tree.at(index0);
    auto bot_index = node0.left;
    auto top_index = node0.right;


    if constexpr (IsAdditive)
    {
        auto y0 = tree.at(bot_index);

        T right_value;
        auto right = finalize_helper1<false>(next_opened, tree, top, right_value, top_index);

        sum += y0.value.value() * right_value;

        auto additive_left = finalize_helper1<true>(next_opened, tree, bot, sum, bot_index);
        finalize_helper2(tree, top, sum, right, node0.additive_right);
        return add_to_tree(tree, std::nullopt, IsAdditive, -1, right, additive_left, -1);
    }
    else
    {
        T left_value;
        auto left = finalize_helper1<false>(next_opened, tree, bot, left_value, bot_index);

        auto y1 = tree.at(top_index);

        sum += left_value * y1.value.value();

        finalize_helper2(tree, bot, sum, left, node0.additive_left);
        auto additive_right = finalize_helper1<true>(next_opened, tree, top, sum, top_index);
        return add_to_tree(tree, sum + next_opened(), IsAdditive, left, -1, -1, additive_right);
    }
}

template<bool IsTop, typename T>
int finalize_helper0(auto next_opened, helper_tree<T>& tree, std::span<int const> indices)
{
    if (indices.size() <= MAX_BASECASE_SIZE)
    {
        assert(not IsTop); // TODO: support small products
        auto opened = next_opened();
        return add_to_tree(tree, opened, false);
    }

    auto [bot, top] = split_indices(indices);
    auto left = finalize_helper0<false>(next_opened, tree, bot);
    auto right = finalize_helper0<false>(next_opened, tree, top);

    T value = tree[left].value.value() * tree[right].value.value();

    auto additive_left = finalize_helper1<true>(next_opened, tree, bot, value, left);
    auto additive_right = finalize_helper1<true>(next_opened, tree, top, value, right);

    if constexpr (not IsTop)
    {
        value += next_opened();
    }

    return add_to_tree(tree, value, false, left, right, additive_left, additive_right);
}

template<class T>
T Beaver<T>::finalize_product(std::size_t factor_count)
{
    std::vector<int> indices(factor_count);
    std::iota(begin(indices), end(indices), 0);

    auto next_opened = [this]()
    {
        auto value = *it;
        ++it;
        return value;
    };

    using tree_t = helper_tree<typename T::open_type>;

    tree_t tree;

    auto index = finalize_helper0<true>(next_opened, tree, indices);


    T last_unopened_element; // TODO: properly store/load this value
    return T::constant(tree[index].value.value(), P.my_num(), MC->get_alphai()) + last_unopened_element;
}

template<typename T>
struct helper_for_prefix_product
{
    T value;
    std::span<int const> indices;
};

template<bool IsLeft, typename T>
void finalize_helper_prefix_fill_intermediate(T one, register_span<T> result, helper_tree<typename T::open_type> const& tree, int index, std::span<int const> indices, std::vector<helper_for_prefix_product<typename T::open_type>>& y, auto opened, std::vector<T>& tuple)
{
    auto [bot, top] = split_indices(indices);
    auto node = tree.at(index);

    if (indices.size() <= MAX_BASECASE_SIZE)
    {
        auto tuple_base = tuple.size();
        for (std::size_t i = 0; i < indices.size() - 1; ++i)
        {
            auto last_tuple_size = tuple.size() - tuple_base;
            y.emplace_back(opened[indices[i]], indices.subspan(i, 1));
            add_tuple<false>(
                one, 
                [&y](std::size_t i){ return y[i].value; }, 
                [&tuple, tuple_base, last_tuple_size](std::size_t i)
                {
                    i = i - 1; // because of the way add_tuple<false> works
                    if (i < last_tuple_size)
                    {
                        return tuple.at(tuple_base - 1 + i);
                    }
                    else
                    {
                        return next_tuple_element(tuple); 
                    }
                }, 
                y.size(), 
                result[indices[i]] = T{});
        }
    }
    else
    {
        finalize_helper_prefix_fill_intermediate<true>(one, result, tree, node.left, bot, y, opened, tuple);
        finalize_helper_prefix_fill_intermediate<false>(one, result, tree, node.right, top, y, opened, tuple);
    }

    if constexpr (IsLeft)
    {
        while (true)
        {
            auto& last = y.back();
            if (indices.front() <= last.indices.back())
            {
                y.pop_back();
            }
            else
            {
                break;
            }
        }

        auto reusable_factor_count = y.size();
        y.emplace_back(node.value.value(), indices);

        auto last_tuple_size = ::pow2(reusable_factor_count) - 1;
        add_tuple<false>(
            one, 
            [&y](std::size_t i){ return y[i].value; }, 
            [&tuple, last_tuple_size](std::size_t i)
            {
                i = i - 1; // because of the way add_tuple<false> works
                if (i < last_tuple_size)
                {
                    return tuple.at(tuple.size() - last_tuple_size + i);

                }
                else
                {
                    return next_tuple_element(tuple); 
                }
            }, 
            y.size(),
            result[indices.back()] = T{});
    }
}

template<bool IsTop, typename T>
void finalize_helper_prefix_fill(register_span<T> result, helper_tree<typename T::open_type> const& tree, int index, std::span<int const> indices, auto opened, std::vector<T>& tuple, auto as_constant)
{
    if (indices.size() <= MAX_BASECASE_SIZE)
    {
        std::vector<typename T::open_type> y;

        assert(indices.size() > 0);
        assert(indices[0] == 0);
        y.reserve(indices.size() - 1);

        y.push_back(opened[0]);
        for (std::size_t i = 1 /* result[0] is already set */; i < indices.size() - 1; ++i)
        {
            assert(static_cast<std::size_t>(indices[i]) == i);
            y.push_back(opened[i]);
            add_tuple<false>(as_constant(1), [&y](std::size_t i){ return y[i]; }, [&tuple](std::size_t){ return next_tuple_element(tuple); }, y.size(), result[i] = T{});
        }
        assert(static_cast<std::size_t>(indices.back()) == indices.size() - 1);
        result[indices.back()] = as_constant(tree.at(index).value.value()) + next_tuple_element(tuple);
        return;
    }

    auto [bot, top] = split_indices(indices);
    auto node = tree.at(index);

    finalize_helper_prefix_fill<false>(result, tree, node.left, bot, opened, tuple, as_constant);

    std::vector<helper_for_prefix_product<typename T::open_type>> y;
    y.emplace_back(tree[node.left].value.value(), bot);
    finalize_helper_prefix_fill_intermediate<false>(as_constant(1), result, tree, node.right, top, y, opened, tuple);

    if constexpr (IsTop)
    {
        T last_unopened_element; // TODO: properly store/load this value
        result[indices.back()] = as_constant(tree[index].value.value()) + last_unopened_element;
    }
    else
    {
        result[indices.back()] = as_constant(tree.at(index).value.value()) + next_tuple_element(tuple);
    }
}

template<typename T>
void Beaver<T>::finalize_prefix_product(register_span<T> result)
{
    auto factor_count = result.size();

    std::vector<int> indices(factor_count);
    std::iota(begin(indices), end(indices), 0);

    auto next_opened = [this]()
    {
        auto value = *it;
        ++it;
        return value;
    };

    using tree_t = helper_tree<typename T::open_type>;

    tree_t tree;

    auto index = finalize_helper0<true>(next_opened, tree, indices);

    std::vector<T> tuple;
    finalize_helper_prefix_fill<true>(result, tree, index, indices, previous_it, tuple, [this](auto value) -> T { return T::constant(value, P.my_num(), MC->get_alphai()); });

    previous_it += factor_count;
}

template <class T>
void Beaver<T>::prepare_monomial(register_span<T const> variables)
{
    auto masks = std::vector<T>(variables.size(), T{});

    shares.reserve(shares.size() + variables.size());

    for (std::size_t i = 0; i < variables.size(); ++i)
    {
        shares.push_back(variables[i] - masks[i]);
    }
}

template <class T>
T Beaver<T>::finalize_monomial(std::span<int const> exponents)
{
    auto max_exponent = *std::max_element(exponents.begin(), exponents.end());
    adjust_binomial_coefficients(max_exponent);

    auto variable_count = exponents.size();
    auto sub_exponents = std::vector<int>(variable_count, 0);
    auto exponent_count = std::transform_reduce(exponents.begin(), exponents.end(), 1, std::multiplies<int>{}, 
                                                [](int exponent){ return exponent + 1; });

    /* TODO: Set tuple to the data generated in the offline phase 
    *
    * Assumption: tuple is of the following form:
    * (1, a, a^2, ..., a^(d_1), b, ab, a^2b, ..., a^(d_1)b, ..., a^(d_1)b^(d_2), ...)
    * In particular, special care has to be taken when calling this method from finalize_polynomial
    * since the tuple entries must be duplicated during the offline phase. */
    auto tuple = std::vector<T>(exponent_count, T{});
    tuple[0] = T::constant(1, P.my_num(), MC->get_alphai());

    auto next_exponents = [&sub_exponents, exponents]()
    {
        for (std::size_t i = 0; i < sub_exponents.size(); ++i)
        {
            if (++sub_exponents[i] > exponents[i])
            {
                sub_exponents[i] = 0;
            }
            else
            {
                return true;
            }
        }
        return false;
    };

    auto result = T{};
    auto tuple_iterator = tuple.begin();

    do
    {
        assert(tuple_iterator != tuple.end());

        auto product = *tuple_iterator;
        for (std::size_t k = 0; k < variable_count; ++k)
        {
            auto d = exponents[k];
            auto e = sub_exponents[k];
            auto y = it[k];

            product *= typename T::clear{binomial_coefficients[d][e]} * ::pow(y, d - e);
        }
        result += product;
        ++tuple_iterator;
    } while (next_exponents());

    assert(tuple_iterator == tuple.end());

    it += variable_count;

    return result;
}

template <class T>
void Beaver<T>::prepare_polynomial(register_span<T const> variables)
{
    // Without offline phase, this method is identical to prepare_monomial.
    // Later, the mask elements have to access different elements of the offline tuple.
    auto masks = std::vector<T>(variables.size(), T{});

    shares.reserve(shares.size() + variables.size());

    for (std::size_t i = 0; i < variables.size(); ++i)
    {
        shares.push_back(variables[i] - masks[i]);
    }
}

template<class T>
T Beaver<T>::finalize_polynomial(register_span<typename T::clear const> coefficients, std::span<int const> exponents)
{
    auto variable_count = exponents.size() / coefficients.size();

    // adjust binomial coefficients only once
    auto max_exponent = *std::max_element(exponents.begin(), exponents.end());
    adjust_binomial_coefficients(max_exponent);

    auto result = T{};
    for (std::size_t i = 0; i < coefficients.size(); ++i)
    {
        result += coefficients[i] * finalize_monomial(exponents.subspan(i * variable_count, variable_count));
        it -= variable_count;
    }
    it += variable_count;

    return result;
}


template<class T>
void Beaver<T>::exchange()
{
    MC->POpen(opened, shares, P);
    it = opened.begin();
    triple = triples.begin();

    shares.clear();
}

template<class T>
void Beaver<T>::second_exchange()
{
    assert(it == opened.end());

    previously_opened.clear();
    std::swap(opened, previously_opened);
    previous_it = previously_opened.begin();

    exchange();
}

template<class T>
void Beaver<T>::start_exchange()
{
    MC->POpen_Begin(opened, shares, P);
}

template<class T>
void Beaver<T>::stop_exchange()
{
    MC->POpen_End(opened, shares, P);
    it = opened.begin();
    triple = triples.begin();

    shares.clear();
}

template<class T>
T Beaver<T>::finalize_mul(int n)
{
    (void) n;
    typename T::open_type masked[2];
    T& tmp = (*triple)[2];
    for (int k = 0; k < 2; k++)
    {
        masked[k] = *it++;
    }
    tmp += (masked[0] * (*triple)[1]);
    tmp += ((*triple)[0] * masked[1]);
    tmp += T::constant(masked[0] * masked[1], P.my_num(), MC->get_alphai());
    triple++;
    return tmp;
}

#endif
