#!/usr/bin/env python3

import sys
import os
import shlex
import fire
import subprocess
import itertools
import matplotlib
import json
import ast
from datetime import datetime
from tqdm import tqdm as progressbar
from tqdm import trange as progress_range

# Takes a program (in our case to compute a multivariate polynomial with polytuples), runs benchmarks for different network delays and returns the output as plots.
 
def main(program, *args, output_folder="benchmarks", approaches=("base", "tuple"), delays=[0,5,10,20], repeats=10, load_json=False, save_json=True, save_md=True, save_plot=True, plot_x_parameter=None, labels=[], plot_time_title="", plot_data_title="", with_data=False, scale=0.5):
    flat_parameters = []
    ranged_parameters = []
    for i, arg in enumerate(args):
        if isinstance(arg, tuple):
            flat_parameters.append(list(range(*arg)))
            ranged_parameters.append(i)
        elif isinstance(arg, list):
            flat_parameters.append(arg)
            ranged_parameters.append(i)
        else:
            flat_parameters.append([arg])
    parameters = list(itertools.product(*flat_parameters))

    program_name, _ = os.path.splitext(os.path.basename(program))

    now = datetime.now()
    now = f"{now:%Y-%m-%d_%H-%M-%S}"
    if save_json and isinstance(save_json, bool):
        os.makedirs(output_folder, exist_ok=True)
        save_json = os.path.join(output_folder, f"{program_name}_{now}.json")

    if save_md and isinstance(save_md, bool):
        os.makedirs(output_folder, exist_ok=True)
        save_md = os.path.join(output_folder, f"{program_name}_{now}.md")

    if save_plot and isinstance(save_plot, bool):
        os.makedirs(output_folder, exist_ok=True)
        save_plot = os.path.join(output_folder, f"{program_name}_{now}.pdf")

    if load_json:
        with open(load_json) as f:
            json_result = json.load(f)

        results = {}
        for k, v in json_result.items():
            k = ast.literal_eval(k)
            assert isinstance(k, tuple)
            results[k] = {}
            for kk, vv in v.items():
                try:
                    kk = int(kk)
                except:
                    pass
                results[k][kk] = vv
    else:
        results = {}

        with progressbar(total=len(approaches) * len(parameters) * len(delays) * repeats, desc="total%", nrows=4, position=0) as progress:
            for approach in approaches:
                for parameter in progressbar(parameters, desc="params", position=1, leave=False):
                    experiment = tuple(itertools.chain([program_name, approach], parameter))

                    compile_call = ["./compile.py", program, approach, *(str(x) for x in parameter)]
                    _ = subprocess.check_output(compile_call)
                    results[experiment] = {}
                    for delay in progressbar(delays, desc="delays", position=2, leave=False):
                        try:
                            if delay:
                                tc_start_call = f"sudo tc qdisc add dev lo root handle 1:0 netem limit 100000 delay {delay}ms".split()
                                subprocess.check_call(tc_start_call)
                            collected_time = []
                            collected_data = []
                            for _ in progress_range(repeats, desc="repeat", position=3, leave=False):
                                run_call = ["Scripts/run-online.sh", "-".join(str(x) for x in experiment)]
                                output = subprocess.check_output(run_call, stderr=subprocess.DEVNULL)
                                output = output.decode().splitlines()
                                time_line = next(line for line in output if line.startswith("Time = "))
                                data_line = next(line for line in output if line.startswith("Data sent = "))
                                time = float(time_line.split()[2])
                                data = float(data_line.split()[3])
                                collected_time.append(time)
                                collected_data.append(data)

                                progress.update()

                            time = sum(collected_time) / repeats
                            data = sum(collected_data) / repeats

                            results[experiment][delay] = dict(time=time, data=data)
                        finally:
                            if delay:
                                tc_end_call = "sudo tc qdisc del dev lo root".split()
                                subprocess.check_call(tc_end_call)

    if save_json:
        json_result = {}
        for k, v in results.items():
            json_result[repr(k)] = v
        with open(save_json, "w") as f:
            f.write(json.dumps(json_result, indent=4))
    if save_md:
        with open(save_md, "w") as f:
            f.write("````\n")
            f.write(" ".join([shlex.quote(str(x)) for x in sys.argv]))
            f.write("\n````\n")

            for j, (experiment, result) in enumerate(results.items()):
                for i, delay in enumerate(delays):
                    if i == 0 and j == 0:
                        line =      "delay | program | approach |"
                        separator = "----- | ------- | -------- |"
                        for j, _ in enumerate(experiment[2:]):
                            try:
                                label = labels[j]
                                sep = "-" * len(label)
                            except IndexError:
                                label = "?"
                                sep = "-"
                            line += f" {label} |"
                            separator += f" {sep} |"
                        line +=      " time | data"
                        separator += " ---- | ----"
                        f.write(f"| {line} |\n| {separator} |\n")

                    line = f"{delay} ms | "
                    for parameter in experiment:
                        line += f"{parameter} | "
                    line += str(result[delay]["time"]) + " s | " + str(result[delay]["data"]) + " MB"
                    f.write(f"| {line} |\n")
    if save_plot:
        if plot_x_parameter is None:
            plot_x_parameter = ranged_parameters[0]

        subplot_count = len(parameters) // len(flat_parameters[plot_x_parameter])

        from matplotlib import pyplot as plt

        plottable_time = {}
        plottable_data = {}

        x = {}

        for delay in delays:
            plottable_time[delay] = {}
            plottable_data[delay] = {}
            for approach in approaches:
                plottable_time[delay][approach] = {}
                plottable_data[delay][approach] = {}
                for parameter in parameters:
                    experiment = tuple(itertools.chain([program_name, approach], parameter))

                    result = results[experiment][delay]
                    time = result["time"]
                    data = result["data"]

                    non_x_parameters = tuple(parameter[:plot_x_parameter] + parameter[plot_x_parameter+1:])
                    x_parameter = parameter[plot_x_parameter]

                    try:
                        t = plottable_time[delay][approach][non_x_parameters]
                        d = plottable_data[delay][approach][non_x_parameters]
                    except:
                        t = {}
                        d = {}
                        plottable_time[delay][approach][non_x_parameters] = t
                        plottable_data[delay][approach][non_x_parameters] = d

                    x[x_parameter] = None

                    t[x_parameter] = time
                    d[x_parameter] = data

        X = list(sorted(x.keys()))


        if with_data:
            fig, axes_ = plt.subplots(subplot_count, (len(delays) + 1), figsize=(16 * (len(delays) + 1) * scale, 9 * subplot_count * scale))
        else:
            fig, axes_ = plt.subplots(subplot_count, len(delays), figsize=(16 * len(delays) * scale, 9 * subplot_count * scale))

        def axes(i, j):
            if subplot_count == 1:
                return axes_[j]
            else:
                return axes_[i, j]

        for j, delay in enumerate(delays):
            for approach in approaches:
                for i, (parameter, time) in enumerate(plottable_time[delay][approach].items()):
                    data = plottable_data[delay][approach][parameter]

                    time = [time[x] for x in X]
                    data = [data[x] for x in X]

                    axes(i, j).plot(X, time, label=approach)
                    axes(i, j).set_ylabel("time / s")
                    axes(i, j).set_title(plot_time_title.format(*parameter, delay))

                    if j == 0 and with_data:
                        axes(i, len(delays)).plot(X, data, label=approach)
                        axes(i, len(delays)).set_ylabel("data / MB")
                        axes(i, len(delays)).set_title(plot_data_title.format(*parameter))

            try:
                axes(subplot_count - 1, j).set_xlabel(labels[plot_x_parameter])
                if with_data:
                    axes(subplot_count - 1, len(delays)).set_xlabel(labels[plot_x_parameter])
            except IndexError:
                pass

        plt.savefig(save_plot, format="pdf", bbox_inches="tight")


if __name__ == "__main__":
    try:
        fire.Fire(main)
    except subprocess.CalledProcessError as e:
        print("stderr:\n", e.stderr)
        print("stdout:\n", e.stdout)
        raise e
