#!/usr/bin/env python3

import os
import fire
import subprocess
import random
from datetime import datetime

def main(source = "Programs/Source/benchmark_net.mpc", options = ["B"], compiler = "compile.py", offline = "./Fake-Offline.x", runner="./Scripts/run-online.sh", parties=2, repeats=10, delays=[0,0.5,1,2,3,4,5,6,7,8,9,10,12,14,16,18,20], with_data=False, scale=0.5):
    time_results = {}
    data_results = {}
    all_options = (options, list(options) + ["tuple"])
    approaches = ("base", "tuple")
    for approach, options in zip(approaches, all_options):
        requirements = {}
        def add_requirement(key, value, unique_postfix=False):
            if value is not None:
                prefix, postfix = value
                if unique_postfix:
                    old_prefix = requirements.get((key, postfix), 0)
                    if prefix > old_prefix:
                        requirements[(key, postfix)] = prefix
                else:
                    old_prefix, _ = requirements.get(key, (0,0))
                    if prefix > old_prefix:
                        requirements[key] = value

        def extract(line, what, execute=None):
            where = line.find(what)
            if where < 0:
                return None
            prefix = line[:where]
            postfix = line[where + len(what):]
            try:
                prefix = prefix.strip()
                prefix = int(prefix)
            except:
                pass
            try:
                postfix = postfix.strip()
                postfix = int(postfix)
            except:
                pass
            if execute is not None:
                execute(prefix, postfix)
            return prefix, postfix

        def parse_line(key, line, what, unique_postfix=False, execute=None):
            add_requirement(key, extract(line, what, execute=execute), unique_postfix=unique_postfix)

        def generate_fake_input(count, player):
            destination = f"Player-Data/Input-P{player}-0"
            os.makedirs(os.path.dirname(destination), exist_ok=True)
            with open(destination, "w") as f:
                for _ in range(count):
                    x = random.uniform(-1.0, 1.0)
                    f.write(str(x))
                    f.write(" ")

        compilation = subprocess.check_output(["python3", compiler, source, *options])
        compilation = compilation.decode()
        for line in compilation.splitlines():
            parse_line("inputs", line, "integer inputs from player", execute=generate_fake_input)
            parse_line("bits", line, "integer bits")
            parse_line("dabits", line, "integer dabits")
            parse_line("triples", line, "integer triples")
            parse_line("bit triples", line, "integer triples")
            parse_line("matrix", line, "matrix multiplications", unique_postfix=True)
            parse_line("edabits", line, "strict edabits of length", unique_postfix=True)

        offline_args = [str(parties)]
        
        if "inputs" in requirements:
            offline_args.append("-inp")
            offline_args.append("1," + str(requirements["inputs"][0]))
        if "bits" in requirements:
            offline_args.append("-bit")
            offline_args.append("1," + str(requirements["bits"][0]))
        if "triples" in requirements:
            offline_args.append("-trip")
            offline_args.append("1," + str(requirements["triples"][0]))
        if "bit triples" in requirements:
            offline_args.append("-btrip")
            offline_args.append("1," + str(requirements["bit triples"][0]))
        default = 1
        edabits = []
        for key, value in requirements.items():
            if key[0] == "edabits":
                edabits.append(key[1])
                if value > default:
                    default = value

        if edabits:
            offline_args.append("-e")
            offline_args.append(",".join(str(length) for length in edabits))
        if default > 1:
            offline_args.append("-d")
            offline_args.append(str(default))

        _ = subprocess.check_output([offline, *offline_args])

        program_name, _ = os.path.splitext(os.path.basename(source))

        program = "-".join([program_name, *options])

        time_results[approach] = []
        data_results[approach] = []

        for delay in delays:
            try:
                if delay:
                    tc_start_call = f"sudo tc qdisc add dev lo root handle 1:0 netem limit 100000 delay {delay}ms".split()
                    subprocess.check_call(tc_start_call)

                collected_time = []
                collected_data = []

                for _ in range(repeats):
                    output = subprocess.check_output([runner, program])
                    output = output.decode().splitlines()
                    time_line = next(line for line in output if line.startswith("Time = "))
                    data_line = next(line for line in output if line.startswith("Data sent = "))
                    time = float(time_line.split()[2])
                    data = float(data_line.split()[3])
                    collected_time.append(time)
                    collected_data.append(data)

                time = sum(collected_time) / len(collected_time)
                data = sum(collected_data) / len(collected_data)

                time_results[approach].append(time)
                data_results[approach].append(data)
            finally:
                if delay:
                    tc_end_call = "sudo tc qdisc del dev lo root".split()
                    subprocess.check_call(tc_end_call)

    from matplotlib import pyplot as plt
    
    if with_data:
        fig, axes = plt.subplots(1, 2, figsize=(16*2*scale, 9*scale))
    else:
        fig, axes = plt.subplots(1, 1, figsize=(16*scale, 9*scale))

    for approach in approaches:
        if with_data:
            axes[0].plot(delays, time_results[approach])
            axes[1].plot(delays, data_results[approach])
            axes[0].set_xlabel("delay / ms")
            axes[0].set_ylabel("time / s")
            axes[1].set_xlabel("delay / ms")
            axes[1].set_ylabel("data / MB")
        else:
            axes.plot(delays, time_results[approach])
            axes.set_xlabel("delay / ms")
            axes.set_ylabel("time / s")

    program_name = program_name + "-" + "-".join(all_options[0])

    now = datetime.now()
    now = f"{now:%Y-%m-%d_%H-%M-%S}"
    # if save_json and isinstance(save_json, bool):
    #     os.makedirs(output_folder, exist_ok=True)
    #     save_json = os.path.join(output_folder, f"{program_name}_{now}.json")

    # if save_md and isinstance(save_md, bool):
    #     os.makedirs(output_folder, exist_ok=True)
    #     save_md = os.path.join(output_folder, f"{program_name}_{now}.md")

    # if save_plot and isinstance(save_plot, bool):
    output_folder = "benchmarks"
    os.makedirs(output_folder, exist_ok=True)
    save_plot = os.path.join(output_folder, f"{program_name}_{now}.pdf")
    plt.savefig(save_plot, format="pdf", bbox_inches="tight")

if __name__ == "__main__":
    fire.Fire(main)