#!/usr/bin/env python3

import os
from sys import stderr
from threading import Timer
import fire
import subprocess
from datetime import datetime
from tqdm import tqdm as progressbar
from tqdm import trange as progress_range

# https://stackoverflow.com/questions/12435211/threading-timer-repeat-function-every-n-seconds?noredirect=1&lq=1
class RepeatTimer(Timer):
    def run(self):
        while not self.finished.wait(self.interval):
            self.function(*self.args, **self.kwargs)

def update_progress(progress):
    n = progress.n
    if n + 2 >= progress.total:
        progress.n = 0
    progress.update()

class Spinner():
    def __init__(self, desc, position, interval=0.5):
        self.desc = desc
        self.position = position
        self.interval = interval

    def __enter__(self):
        self.progress = progressbar(total=60/self.interval, desc=self.desc, position=self.position, leave=False, bar_format="{desc}:   ?%|{bar}| ?/? [{elapsed}<?, ?it/s]")
        self.timer = RepeatTimer(self.interval, update_progress, (self.progress,))
        self.timer.start()
        return self

    def __exit__(self, type, value, trace_back):
        self.timer.cancel()
        self.progress.close()

def get_requirements(output):
    requirements = {}
    def add_requirement(key, value, unique_postfix=False):
        if value is not None:
            prefix, postfix = value
            if unique_postfix:
                old_prefix = requirements.get((key, postfix), 0)
                if prefix > old_prefix:
                    requirements[(key, postfix)] = prefix
            else:
                old_prefix, _ = requirements.get(key, (0,0))
                if prefix > old_prefix:
                    requirements[key] = value

    def extract(line, what, execute=None):
        where = line.find(what)
        if where < 0:
            return None
        prefix = line[:where]
        postfix = line[where + len(what):]
        try:
            prefix = prefix.strip()
            prefix = int(prefix)
        except:
            pass
        try:
            postfix = postfix.strip()
            postfix = int(postfix)
        except:
            pass
        if execute is not None:
            execute(prefix, postfix)
        return prefix, postfix

    def parse_line(key, line, what, unique_postfix=False, execute=None):
        add_requirement(key, extract(line, what, execute=execute), unique_postfix=unique_postfix)

    for line in output.splitlines():
        parse_line("inputs", line, "integer inputs from player")
        parse_line("bits", line, "integer bits")
        parse_line("dabits", line, "integer dabits")
        parse_line("triples", line, "integer triples")
        parse_line("bit triples", line, "integer triples")
        parse_line("matrix", line, "matrix multiplications", unique_postfix=True)
        parse_line("edabits", line, "strict edabits of length", unique_postfix=True)

    return requirements

def run_offline(offline, parties, prime, requirements):
    offline_args = [str(parties)]
    if prime:
        offline_args.append("-P")
        offline_args.append(prime)
        
    if "inputs" in requirements:
        offline_args.append("-inp")
        offline_args.append("1," + str(requirements["inputs"][0]))
    if "bits" in requirements:
        offline_args.append("-bit")
        offline_args.append("1," + str(requirements["bits"][0]))
    if "triples" in requirements:
        offline_args.append("-trip")
        offline_args.append("1," + str(requirements["triples"][0]))
    if "bit triples" in requirements:
        offline_args.append("-btrip")
        offline_args.append("1," + str(requirements["bit triples"][0]))
    default = 1
    edabits = []
    for key, value in requirements.items():
        if key[0] == "edabits":
            edabits.append(key[1])
            if value > default:
                default = value

    if edabits:
        offline_args.append("-e")
        offline_args.append(",".join(str(length) for length in edabits))
    if default > 1:
        offline_args.append("-d")
        offline_args.append(str(default))

    _ = subprocess.check_output([offline, *offline_args], stderr=subprocess.PIPE)

def main(pattern = "Programs/Source/ordinos_v{}.mpc", instances = [1,2,3,4], options=[], parties=2, compiler = "compile.py", compiler_options="<prime>", params="Player-Data/2-p-64/Params-Data", offline="./Fake-Offline.x", runner="./Scripts/run-online.sh", repeats=32, delays=[0,0.5,1,2,3,4,5,6,7,8,9,10,12,14,16,18,20], with_data=False, with_core_time=False, scale=0.5):
    time_results = {}
    overalltime_results = {}
    data_results = {}
    all_options = (options, list(options) + ["tuple"], list(options) + ["edabit"])
    approaches = ("base", "tuple", "edabit")
    with progressbar(total=len(approaches) * len(instances) * len(delays) * repeats, desc="total%", nrows=5, position=0) as progress:
        for instance in progressbar(instances, desc="instnc", position=1, leave=False):
            time_results[instance] = {}
            overalltime_results[instance] = {}
            data_results[instance] = {}

            for approach, options in zip(progressbar(approaches, desc="apprch", position=2, leave=False), all_options):
                source = pattern.format(instance)

                compile_options = []
                if "<prime>" in compiler_options:
                    compile_options.append("-P")
                    with open(params, "r") as f:
                        p = f.readline()
                    compile_options.append(p)

                with Spinner(desc="cmpile", position=3):
                    output = subprocess.check_output(["python3", compiler, *compile_options, source, *options])
                    output = output.decode()

                requirements = get_requirements(output)
                with Spinner(desc="offlne", position=3):
                    run_offline(offline, parties, None, requirements)
                    run_offline(offline, parties, p, requirements)

                program_name, _ = os.path.splitext(os.path.basename(source))

                program = "-".join([program_name, *options])

                time_results[instance][approach] = []
                overalltime_results[instance][approach] = []
                data_results[instance][approach] = []

                for delay in progressbar(delays, desc="delays", position=3, leave=False):
                    try:
                        if delay:
                            tc_start_call = f"sudo tc qdisc add dev lo root handle 1:0 netem limit 100000 delay {delay}ms".split()
                            subprocess.check_call(tc_start_call)

                        collected_time = []
                        collected_overalltime = []
                        collected_data = []

                        for _ in progress_range(repeats, desc="repeat", position=4, leave=False):
                            output = subprocess.check_output([runner, program], stderr=subprocess.PIPE)
                            output = output.decode().splitlines()
                            time_line = next(line for line in output if line.startswith("Time1 = "))
                            overalltime_line = next(line for line in output if line.startswith("Time = "))
                            data_line = next(line for line in output if line.startswith("Data sent = "))
                            time = float(time_line.split()[2])
                            overalltime = float(overalltime_line.split()[2])
                            data = float(data_line.split()[3])
                            collected_time.append(time)
                            collected_overalltime.append(overalltime)
                            collected_data.append(data)

                            progress.update()

                        time = sum(collected_time) / len(collected_time)
                        overalltime = sum(collected_overalltime) / len(collected_overalltime)
                        data = sum(collected_data) / len(collected_data)

                        time_results[instance][approach].append(time)
                        overalltime_results[instance][approach].append(overalltime)
                        data_results[instance][approach].append(data)
                    finally:
                        if delay:
                            tc_end_call = "sudo tc qdisc del dev lo root".split()
                            subprocess.check_call(tc_end_call)

    from matplotlib import pyplot as plt

    if with_data:
        fig, axes_ = plt.subplots(len(instances), 2, figsize=(16 * 2 * scale, 9 * len(instances) * scale))
    else:
        fig, axes_ = plt.subplots(len(instances), 1, figsize=(16 * scale, 9 * len(instances) * scale))

    def axes(i, j):
        if with_data:
            if len(instances) > 1:
                return axes_[i, j]
            else:
                return axes_[j]
        else:
            if len(instances) > 1:
                return axes_[i]
            else:
                return axes_


    colors = {}

    for i, instance in enumerate(instances):
        for j, approach in enumerate(approaches):
            if with_core_time:
                line = axes(i, 0).plot(delays, time_results[instance][approach], color=colors.get(j, None))
                colors[j] = line[0].get_color()
                line = axes(i, 0).plot(delays, overalltime_results[instance][approach], color=colors.get(j, None), linestyle="dashed")
            else:
                axes(i, 0).plot(delays, overalltime_results[instance][approach])
            axes(i, 0).set_xlabel("delay / ms")
            axes(i, 0).set_ylabel("time / s")
            # axes(i, 0).set_title("runtime of ordinos_v{}".format(instance))

            if with_data:
                line = axes(i, 1).plot(delays, data_results[instance][approach], color=colors.get(j, None))
                axes(i, 1).set_xlabel("delay / ms")
                axes(i, 1).set_ylabel("data / MB")
                # axes(i, 1).set_title("communication cost of ordinos_v{}".format(instance))



    source = pattern.format("-".join(str(instance) for instance in instances))
    program_name, _ = os.path.splitext(os.path.basename(source))

    now = datetime.now()
    now = f"{now:%Y-%m-%d_%H-%M-%S}"
    # if save_json and isinstance(save_json, bool):
    #     os.makedirs(output_folder, exist_ok=True)
    #     save_json = os.path.join(output_folder, f"{program_name}_{now}.json")

    # if save_md and isinstance(save_md, bool):
    #     os.makedirs(output_folder, exist_ok=True)
    #     save_md = os.path.join(output_folder, f"{program_name}_{now}.md")

    # if save_plot and isinstance(save_plot, bool):
    output_folder = "benchmarks"
    os.makedirs(output_folder, exist_ok=True)
    save_plot = os.path.join(output_folder, f"{program_name}_{now}.pdf")
    plt.savefig(save_plot, format="pdf", bbox_inches="tight")

if __name__ == "__main__":
    fire.Fire(main)